#ifndef CREATURE_HPP
#define CREATURE_HPP

#include <iostream>

namespace ecosystem{
	
	//Forward declarations
	template <typename T>
	class Dynamic2D;
	
	enum class Livingstate : bool{
		ALIVE=true,DEAD=false
	};
	
	//Interfaces
	struct Clonable {
		virtual Clonable* clone() const = 0;
		virtual ~Clonable() {};
	};
	
	//ClonableTestClass *clone() const override{ return new ClonableTestClass(*this); }

	struct Serializable{
		virtual std::ostream& output(std::ostream& os) const =0;
		virtual std::istream& input(std::istream& is) =0;
	};

	//Insertor,Extractor
	inline std::ostream& operator<<(std::ostream& os,const Serializable& serializable){
		return serializable.output(os);
	}
	inline std::istream& operator>>(std::istream& is,Serializable& serializable){
		return serializable.input(is);
	}
	
	

    class Creature: public Serializable, public Clonable{
        private:
		
			Livingstate is_alive=Livingstate::ALIVE;
			
		protected:
		
			std::string name;
            int current_age;
            int max_age;
            int energy;
            int energy_consumption_per_round;
			char symbol=' ';
			wchar_t uni_symbol=L' ';
			
			
			
        public:
			bool updated=false;
            Creature(const std::string& nam,int m_age,int e,int e_per_round, char sym,wchar_t uni_sym);
            
            virtual void move(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y)=0;
            virtual void eat()=0;
			
			virtual void update(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y)=0;
			
			Livingstate get_living_state() const;
			void display() const;
			
			void perish();
			void age();
			
			bool is_printable();
			const char& get_symbol() const;
			const wchar_t& get_unisymbol() const;
			
			void gain_energy(int amm);
			
            virtual ~Creature()=default;
			
			//Overrides:
			std::ostream& output(std::ostream& os) const override;
			std::istream& input(std::istream& is) override;
    };
}
#endif