
#include <iostream>

#include "define_config.hpp"
#include "dynamic_2d.hpp"
#include "creature.hpp"
#include "empty_creature.hpp"
#include "random_gen.hpp"

#include <fstream>
#include <string>

//TODO reorganize classes, currently this is unbearable.

namespace ecosystem{

    Creature::Creature(const std::string& nam,int m_age,int e,int e_per_round, char sym,wchar_t uni_sym): name(nam), max_age(m_age),energy(e),energy_consumption_per_round(e_per_round),symbol(sym),uni_symbol(uni_sym){
		current_age=0;
		if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=4){
			std::cout << "!!! Constructor" << symbol << " called!!!";
		}
		updated=false;
	}
	
	Livingstate Creature::get_living_state() const {
		return is_alive;
	}
	void Creature::display() const {
        std::cout << name << " | Energy: " << energy << " | Consumption per round: " << energy_consumption_per_round << " | Age: " << current_age << " | Exp.max age: " << max_age << "\n";
    }
	
	void Creature::perish(){
		is_alive=Livingstate::DEAD;
	}
	void Creature::age(){
		++current_age;
		energy-=energy_consumption_per_round;
		updated=true;
		if(current_age>max_age||energy<=0){
			perish();
		}
	}
	
	bool Creature::is_printable(){
		if(symbol==' '){
			return false;
		}
		return true;
	}
	const char& Creature::get_symbol() const {
		return symbol;
	}
	const wchar_t& Creature::get_unisymbol() const{
		return uni_symbol;
	}
	
	void Creature::gain_energy(int amm){
		energy+=amm;
		if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
			std::cout << name << " gained " << amm << ". New energy: " << energy << "\n";
		}
	}
	
	//Overrides:
	std::ostream& Creature::output(std::ostream& os) const{
		
		return os << symbol;
	}
	std::istream& Creature::input(std::istream& is){
		char delimiter; //For reading in the |
		return is >> name>>delimiter>>energy>>delimiter>>energy_consumption_per_round>>delimiter>>current_age>>delimiter>>max_age;
	}
	
	class Environmental{
		private:
		
			Livingstate is_lasting=Livingstate::ALIVE;
			
		protected:
			int duration_left=0;
			
			int affect_chance=0; //100% will affect all enemies
			
			int energy_drain_min;
			int energy_drain_max;
			
			int health_min;
			int health_max;
			
			std::string name;
		public:
			Environmental(const std::string& n,int dur_left,int af_chance,int e_d_min, int e_d_max, int h_min, int h_max): duration_left(dur_left), affect_chance(af_chance), energy_drain_min(e_d_min), energy_drain_max(e_d_max),health_min(h_min),health_max(h_max){
				
			}
			
			void delast_self(){
				is_lasting=Livingstate::DEAD;
			}
			
			void age_duration(){
				duration_left--;
				if(duration_left<=0){
					
				}
			}
			virtual void update(const Creature& cur_creature)=0;
			
	};
	

	Empty::Empty() : Creature("Empty tile", RandomGenerator::get_instance().get_random_int(20)+20,100,-5,' ',L'⃣') {}
	
	void Empty::move(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y) {
		//This is empty
	}

	void Empty::eat() {
		//This is empty
	}
	void Empty::update(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y){
		//This is empty
	}
	void Empty::display_self(){
		std::cout << "This is an empty tile!\n";
	}
	Empty* Empty::clone() const{
		return new Empty();
	}

	class Grass : public Creature {
		private:
			std::string surname;
			int days_since_watered;
		public:
			Grass(const std::string& n=" ") : surname(n), Creature("Grass", RandomGenerator::get_instance().get_random_int(20)+20,10,-2,'W',L'෴') {
				int days_since_watered=RandomGenerator::get_instance().get_random_int(5)+5;
			}
			
			void move(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y) override {
				//Grass does not move
			}

			void eat() override {
				gain_energy(5);
				
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=2){
					std::cout << name << " got enough water to thrive.\n";
				}
			}
			bool eat_if_can(){
				if(days_since_watered>0){
					eat();
					days_since_watered--;
					return true;
				}
				
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=2){
					std::cout << name << " did not get newly watered.\n";
				}
				
				return false;
			}
			void update(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y) override {
				if(get_living_state()==Livingstate::ALIVE){
					age();
					eat_if_can();
				}
			}
			void display_self(){
				if(surname!=" "){
					std::cout << "(" << surname << ") ";
				}
				display();
			}
			
			Grass* clone() const override{
				return new Grass(surname);
			}
	};
	
	class Pig : public Creature {
		private:
			std::string surname;
		public:
			Pig(const std::string& n=" ") : surname(n), Creature("Pig", RandomGenerator::get_instance().get_random_int(20)+20,100,-5,'@',L'🐷') {}
			
			void move(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y) override {
				std::vector<std::pair<int, int>> directions;
				if(dynamic_world.is_valid_coordinate(global_x-1,global_y)==true&&dynamic_world[global_y][global_x-1]->get_symbol()==' '){ //Left
					directions.push_back({global_x - 1, global_y});
				}
				if(dynamic_world.is_valid_coordinate(global_x+1,global_y)==true&&dynamic_world[global_y][global_x+1]->get_symbol()==' '){ //Right
					directions.push_back({global_x + 1, global_y});
				}
				if(dynamic_world.is_valid_coordinate(global_x,global_y-1)==true&&dynamic_world[global_y-1][global_x]->get_symbol()==' '){ //Up
					directions.push_back({global_x, global_y-1});
				}
				if(dynamic_world.is_valid_coordinate(global_x,global_y+1)==true&&dynamic_world[global_y+1][global_x]->get_symbol()==' '){ //Down
					directions.push_back({global_x, global_y+1});
				}
				
				if (directions.empty()==true){
					if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
						std::cout << name << " did not have a space to move.\n";
					}
					return;
				}
				
				int rand_idx=RandomGenerator::get_instance().get_random_int(directions.size());
				
				int new_x = directions[rand_idx].first;
				int new_y = directions[rand_idx].second;
				
				dynamic_world.swap_at(global_x,global_y,new_x,new_y);
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
					std::cout << name << " moves cautiously.\n";
				}
			}

			void eat() override {
				gain_energy(12);
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=2){
					std::cout << name << " eaten a grass or a corpse.\n";
				}
			}
			bool eat_if_can(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y){
				std::vector<std::pair<int, int>> directions;
				if(dynamic_world.is_valid_coordinate(global_x-1,global_y)==true&&(dynamic_world[global_y][global_x-1]->get_symbol()=='W'||dynamic_world[global_y][global_x-1]->get_living_state()==Livingstate::DEAD)){ //Left
					directions.push_back({global_x - 1, global_y});
				}
				if(dynamic_world.is_valid_coordinate(global_x+1,global_y)==true&&(dynamic_world[global_y][global_x+1]->get_symbol()=='W'||dynamic_world[global_y][global_x+1]->get_living_state()==Livingstate::DEAD)){ //Right
					directions.push_back({global_x + 1, global_y});
				}
				if(dynamic_world.is_valid_coordinate(global_x,global_y-1)==true&&(dynamic_world[global_y-1][global_x]->get_symbol()=='W'||dynamic_world[global_y-1][global_x]->get_living_state()==Livingstate::DEAD)){ //Up
					directions.push_back({global_x, global_y-1});
				}
				if(dynamic_world.is_valid_coordinate(global_x,global_y+1)==true&&(dynamic_world[global_y+1][global_x]->get_symbol()=='W'||dynamic_world[global_y+1][global_x]->get_living_state()==Livingstate::DEAD)){ //Down
					directions.push_back({global_x, global_y+1});
				}
				
				if (directions.empty()==true){
					if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
						std::cout << name << " did not have anything to eat.\n";
					}
					return false;
				}
				
				int rand_idx=RandomGenerator::get_instance().get_random_int(directions.size());
				
				int new_x = directions[rand_idx].first;
				int new_y = directions[rand_idx].second;
				
				dynamic_world[new_y][new_x]->perish();
				eat();
				return true;
			}
			void update(Dynamic2D<Creature*>& dynamic_world,int global_x,int global_y) override{
				if(get_living_state()==Livingstate::ALIVE){
					age();
					if(eat_if_can(dynamic_world,global_x,global_y)==false){
						move(dynamic_world,global_x,global_y);
					}
				}
			}
			void display_self(){
				if(surname!=" "){
					std::cout << "(" << surname << ") ";
				}
				display();
			}
			
			Pig* clone() const override{
				return new Pig(surname);
			}
	};
	
	//Custom errors
	struct CreatureOutOfIndexError : std::runtime_error {
		CreatureOutOfIndexError(const std::string& what) : std::runtime_error(what) {}
	};
	
	struct EnvironmentalOutOfIndexError : std::runtime_error {
		EnvironmentalOutOfIndexError(const std::string& what) : std::runtime_error(what) {}
	};
	
	template<typename T, typename Container = std::vector<T*>, class ErrorClass = std::out_of_range>
	class CreatureHeteroStore {
		private:
			Container data_container;

		public:
			CreatureHeteroStore() = default;
			
			using const_iterator = typename Container::const_iterator;
			using iterator = typename Container::iterator;
			
			CreatureHeteroStore(const CreatureHeteroStore& other) {
				clear();
				for (const T* obj : other.data_container) {
					Creature* cloned = obj->clone();
						
					T* Tcloned = dynamic_cast<T*>(cloned);
					data_container.push_back(Tcloned);

				}
			}
			CreatureHeteroStore& operator=(const CreatureHeteroStore& other) {
				if (this != &other) {
					clear();

					for (const T* obj : other.data_container) {
						Creature* cloned = obj->clone();
							
						T* Tcloned = dynamic_cast<T*>(cloned);
						data_container.push_back(Tcloned);

					}
				}
				return *this;
			}

			
			const_iterator cbegin() const {
				return data_container.cbegin();
			}
			const_iterator cend() const {
				return data_container.cend();
			}
			
			iterator begin() const {
				return data_container.begin();
			}
			iterator end() const {
				return data_container.end();
			}
			
			size_t size() const{
				return data_container.size();
			}
			void add(T* what){
				try {
					data_container.push_back(what);
				} catch (const std::bad_alloc&) {
					delete what;
					throw ErrorClass("Error: Failed to add new creature to the world queue!");
				}
			}
			T* get_next(){
				T* next = data_container.back();
				
				T* last_clone=static_cast<T*>(next->clone());
				data_container.pop_back();
				
				return last_clone;
			}
			
			void clear(){
				for (T* obj : data_container) {
					delete obj;
				}
				data_container.clear();
			}
				
			~CreatureHeteroStore() {
				clear();
			}
	};
	
	template<typename T, typename Container = std::vector<T*>, class ErrorClass = std::out_of_range>
	class EnvironmentalHeteroStore {
		private:
			Container data_container;

		public:
			EnvironmentalHeteroStore() = default;
			
			using const_iterator = typename Container::const_iterator;
			using iterator = typename Container::iterator;

			//EnvironmentalHeteroStore(const Container& container) : data_container(container) {}
			
			EnvironmentalHeteroStore(const EnvironmentalHeteroStore& other)=delete;
			EnvironmentalHeteroStore& operator=(const EnvironmentalHeteroStore& other) = delete;

			
			const_iterator cbegin() const {
				return data_container.cbegin();
			}
			const_iterator cend() const {
				return data_container.cend();
			}
			
			iterator begin() const {
				return data_container.begin();
			}
			iterator end() const {
				return data_container.end();
			}
			
			size_t size() const{
				return data_container.size();
			}
			
			void add(T* what){
				try {
					data_container.push_back(what);
				} catch (const std::bad_alloc&) {
					delete what;
					throw ErrorClass("Error: Failed to add new environmental effect to the world queue!");
				}
			}
			void erase_at(int index){
				data_container.erase( data_container.begin() + index );
			}
			
			void clear(){
				for (T* obj : data_container) {
					delete obj;
				}
				data_container.clear();
			}
				
			~EnvironmentalHeteroStore() {
				clear();
			}
	};

	class World{
		private:
			CreatureHeteroStore<Creature, std::vector<Creature*>, CreatureOutOfIndexError> creatures_waiting; //Creatures waiting to be spawned, each turn will spawn a new animal
			EnvironmentalHeteroStore<Environmental, std::vector<Environmental*>,EnvironmentalOutOfIndexError> environmental_effects; //A list of environmental_effects
			
			std::size_t world_width=0;
			std::size_t world_height=0;
			
			int current_world_age=0;
			
			Dynamic2D<Creature*> world_space;
			
		public:
			World(std::size_t w,std::size_t h): world_width(w),world_height(h){
				world_space.resize(world_width,world_height,new Empty());
				
				populate_flora();
				populate_fauna();
			}
			void add_creature_to_queue(Creature* new_creature){
				creatures_waiting.add(new_creature);
			}
			bool spawn_next_creature_in_queue(){
				if(creatures_waiting.size()==0){
					return false;
				}
				
				Creature* cur_creature=creatures_waiting.get_next();
				
				std::size_t rX=RandomGenerator::get_instance().get_random_int(world_width);
				std::size_t rY=RandomGenerator::get_instance().get_random_int(world_height);
				while(world_space[rY][rX]->is_printable()==true){
					rX=RandomGenerator::get_instance().get_random_int(world_width);
					rY=RandomGenerator::get_instance().get_random_int(world_height);
				}
				
				
				
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=2){
					std::cout << "Spawned creature: " << *cur_creature << " at X: " << rX << " Y: " << rY << ".\n";
				}
				world_space[rY][rX]=cur_creature;
				return true;
			}
			
			void populate_flora(){
				for(int i=0;i<(world_height+world_width)*2;i++){
					this->add_creature_to_queue(new Grass());
				}
			}
			void populate_fauna(){
				this->add_creature_to_queue(new Pig("Piggy"));
				this->add_creature_to_queue(new Pig("Piggy2"));
			}
			
			void display_world(){
				world_space.print();
			}
			
			void age_world(){
				
				world_space.reset_states();
				world_space.update_loop();
				
				current_world_age++;
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=3){
					std::cout << "The world has aged, now it is " << current_world_age << " years old!\n";
				}
			}
			
			//1 tick of world updates
			void update(){
				
				//Spawn new creatures if they are in the queue.
				bool success=true;
				while(success==true){
					if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=2){
						std::cout << "Trying to spawn new creature...\n";
					}
					success=spawn_next_creature_in_queue();
				}
				
				age_world();
				
				
				if(EcoSystemGlobalConfigManagerSingletonClassInstance::get_instance().get_config_level()>=1){
					display_world();
				}
			}
			void tick_n_times(int n){
				for(int i=0;i<n;i++){
					update();
				}
			}
			void hard_reset_world(){
				//creatures_waiting.clear();
				//environmental_effects.clear();
				//world_space.clear();
			}
			
			/*void read_from_file(const std::string& path){
				hard_reset_world();
				std::ifstream filestream(path);
				if(filestream.is_open()==false){
					std::cerr << "Failed to open file!";
				}
				std::string temp_string;
				
				int cur_line_index=0;
				while (getline(filestream, temp_string)){
					if(cur_line_index==0){
						std::string width="";
						std::string height="";
						bool is_width=true;
						
						for(std::size_t i=0;i<temp_string.size();i++){
							if(temp_string[i]==' '){
								is_width=false;
							} else if(is_width==true){
								width+=temp_string[i];
							} else {
								height+=temp_string[i];
							}
						}
						
						world_width=std::stoi(width);
						world_height=std::stoi(height);
						
						world_space.resize(world_width,world_height,new Empty());
						
					} else {
						for(std::size_t i=0;i<world_width;i++){
							switch(temp_string[i]){
								case 'W':
									world_space.replace_at(i,cur_line_index-1,new Grass());
									break;
								case '@':
									world_space.replace_at(i,cur_line_index-1,new Pig());
									break;
								case ' ':
									world_space.replace_at(i,cur_line_index-1,new Empty());
									break;
							}
						}
					}
					cur_line_index++;
				}
				
				filestream.close();
			}
			void save_to_file(std::string path){
				
			}*/
			
			~World() {
				hard_reset_world();
			}
	};
	
}

int main() {
	ecosystem::World world1(10,10);
	//world1.read_from_file("input_world.dat");
	
	world1.update();
	
	//ecosystem::Pig pig1("Piggy");
	//pig1.update();
	//pig1.display_self();
	//std::cout << pig1;

    return 0;
}