#ifndef RND_GEN_HPP
#define RND_GEN_HPP

#include <iostream>
#include <chrono>
#include <random>
#include <mutex>

namespace ecosystem{
	class EcoSystemGlobalConfigManagerSingletonClassInstance {
		private:
			int debug_level=3;
			std::mutex mtx;

			EcoSystemGlobalConfigManagerSingletonClassInstance(){}

		public:
			EcoSystemGlobalConfigManagerSingletonClassInstance(const EcoSystemGlobalConfigManagerSingletonClassInstance&) = delete;
			
			EcoSystemGlobalConfigManagerSingletonClassInstance& operator=(const EcoSystemGlobalConfigManagerSingletonClassInstance&) = delete;

			static EcoSystemGlobalConfigManagerSingletonClassInstance& get_instance() {
				static EcoSystemGlobalConfigManagerSingletonClassInstance instance;
				return instance;
			}

			int get_config_level() const {
				return debug_level;
			}
			void set_config_level(int n_flag) {
				
				std::lock_guard<std::mutex> lock(mtx);
				
				debug_level=n_flag;
			}
	};
	
	class RandomGenerator {
		private:
			std::random_device rd;
			std::mt19937 gen;
			std::mutex mtx;

			// I swear this is more like lisp than C++
			RandomGenerator() : gen(rd() ^ ((std::mt19937::result_type) std::chrono::duration_cast<std::chrono::seconds>( std::chrono::system_clock::now().time_since_epoch() ).count() + (std::mt19937::result_type) std::chrono::duration_cast<std::chrono::microseconds>( std::chrono::high_resolution_clock::now().time_since_epoch() ).count())) {}

		public:
			RandomGenerator(const RandomGenerator&) = delete;
			
			RandomGenerator& operator=(const RandomGenerator&) = delete;

			static RandomGenerator& get_instance() {
				static RandomGenerator instance;
				return instance;
			}

			int get_random_int(int max) {
				
				std::lock_guard<std::mutex> lock(mtx); //Thread safety, not really needed but looks good
				
				std::uniform_int_distribution<int> distrib(0, max-1);
				return distrib(gen);
			}
	};
}

#endif